//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
/* globals CommuatorState: true  */
function Position(name,comment) {
    this._name = name;
    this._comment = comment;
    this._state = false;
    this._listeners = [];
}

Position.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
    },
    _fireListeners: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i]();
        }
    },
    getName: function() { return this._name; },
    getState: function() { return this._state; },
    setState: function(state) { this._state = state; this._fireListeners(); }
};

function Eastleigh() {
    this._zw30 = false;
    this._next = false;
    this._pointPulled = false;
    this._listeners = [];
}
Eastleigh.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
    },
    _fireListeners: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i]();
        }
    },
    getZw30 : function () { return this._zw30; },
    getZw30Next : function() { return this._next; },
    getPointPulled: function() { return this._pointPulled; },
    setSignals: function(zw30,next) {
        this._zw30 = zw30;
        this._next = next;
        this._fireListeners();
    },
    setPoint: function(pulled) { this._pointPulled = pulled; this._fireListeners(); },
};

var CommuatorState = {
    TRAIN_ON_LINE: 51,
    NORMAL: 144,
    LINE_CLEAR: 224
};

function Commuator(name) {
    this._name = name;
    this._state = CommuatorState.NORMAL;
    this._listeners = [];
}

Commuator.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
    },
    _fireListeners: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i]();
        }
    },
    getName: function() { return this._name; },
    getState: function() { return this._state; },
    setState: function(state) { this._state = state; this._fireListeners(); }
};

function Lever(name,initially) {
    this._pulled = initially;
    this._lockCause = null;
    this._name = name;
    this._listeners = [];
}
Lever.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
    },
    _fireListeners: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i]();
        }
    },
    getName: function() { return this._name; },

    isLocked: function() { return this._lockCause !== null; },
    getLockCause: function() { return this._lockCause ? this._lockCause : ""; },
    unlock: function() { this._lockCause = null; }, // Doesn't fire the listeners since it's called from computeLocks (which is called via a listener itself)
    setLocked: function(cause) { this._lockCause = cause; }, // Doesn't fire the listeners since it's called from computeLocks (which is called via a listener itself)

    isPulled: function() { return this._pulled; },
    pull: function() { if (!this.isLocked()) { this._pulled = true; this._fireListeners(); }},
    push: function() { if (!this.isLocked()) { this._pulled = false; this._fireListeners(); }},
};

function RomseyBoxModel() {
    var computeLockListener = function(us) {
        return function() {
            us.computeLocks();
        };
    }(this);
    
    this._redbridgeUp   = new Commuator("Redbridge Up Line");
    this._redbridgeDown = new Commuator("Redbridge Down Line");
    this._kimbridgeUp   = new Commuator("Kimbridge Up Line");
    this._kimbridgeDown = new Commuator("Kimbridge Down Line");
    this._eastleigh = new Eastleigh();
    this._lever = [];
    this._barrierState = false;
    for(var i=1; i<=23; i++) {
        this._lever[i] = new Lever(i, i === 9); // Only lever 9 is pulled initially
        this._lever[i].addListener(computeLockListener);
    }
    this._free17 = false;
    this._positions = {
            PC : new Position("PC","In front of signal ZW30"),
            PD : new Position("PD","In front of signal 1R"),
            PE : new Position("PE","Between Halterworth Crossing and PF"), 
            PF : new Position("PF","In front of Signal 1"),
            PG : new Position("PG","In front of Points 11"),
            PH : new Position("PH","In front of Signal 2"),
            UZ : new Position("UZ","In front of Signal 17"),
            C  : new Position("C","Immediately after Points 8 on up main"),
            J  : new Position("J","In front of Points 7"),
            D  : new Position("D","In front of Signal 13"),
            H  : new Position("H","After signal 5"),
            K  : new Position("K","In front of Points 10 on down main"),
            L  : new Position("L","Between Points 10 and Platform 2"),
            B  : new Position("B","Between Up siding 2 and Crossover Points West"),
            A  : new Position("A","In front of signal 20"),
            P1 : new Position("","At Platform 1 (up line)"),   // Added to ease use of simulation
            P2 : new Position("","At Platform 2 (down line")   // Added to ease use of simulation 
    };
    this.computeLocks();
    this._listeners = [];
    
    // And add computeLocks as listeners for the commuators too
    this._redbridgeUp.addListener(computeLockListener);
    this._redbridgeDown.addListener(computeLockListener);
    this._kimbridgeUp.addListener(computeLockListener);
    this._kimbridgeDown.addListener(computeLockListener);
}

RomseyBoxModel.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
        for(var i=1; i<this._lever.length; i++) {
            this._lever[i].addListener(listener);
        }
        for(var index in this._positions) {
            if (this._positions.hasOwnProperty(index)) {
                this._positions[index].addListener(listener);
            }
        }
        this._redbridgeUp.addListener(listener);
        this._redbridgeDown.addListener(listener);
        this._kimbridgeUp.addListener(listener);
        this._kimbridgeDown.addListener(listener);
        this._eastleigh.addListener(listener);
    },
    _fireListeners : function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i]();
        }
    },
    getLever : function(i) { return this._lever[i]; },
    getRedbridgeUpCommuator   : function() { return this._redbridgeUp; },
    getRedbridgeDownCommuator : function() { return this._redbridgeDown; },
    getKimbridgeUpCommuator   : function() { return this._kimbridgeUp; },
    getKimbridgeDownCommuator : function() { return this._kimbridgeDown; },
    getEastleigh : function() { return this._eastleigh; },
    is17Free: function() { return this._free17; },
    release17: function() { this._free17 = true; this.computeLocks(); this._fireListeners(); },
    lock17: function() { this._free17 = false; this.computeLocks(); this._fireListeners(); },
    getPosition: function(name) { return this._positions[name]; },
    getBarrierState: function() { return this._barrierState; },
    setBarrierState: function(state) { this._barrierState = state; this.computeLocks(); this._fireListeners(); },
    computeLocks : function() {
        var locks = function(lever, target) {
            if (lever.isPulled()) { target.setLocked("Locked by "+lever.getName()+" being pulled"); } 
        };
        var releasedBySingle = function(target,lever) {
            if (!!!lever.isPulled()) {target.setLocked("Locked by "+lever.getName()+" being unpulled"); }
            locks(target,lever);
        };
        var releasedByBoth = function(target, lever1, lever2) {
            if (!!!lever1.isPulled() && !!!lever2.isPulled()) { target.setLocked("Locked by both "+lever1.getName()+" and "+lever2.getName() +" being unpulled. Both must be pulled to release this lever"); }
            else if (!!!lever1.isPulled()) { target.setLocked("Locked by "+lever1.getName()+" being unpulled"); }
            else if (!!!lever2.isPulled()) { target.setLocked("Locked by "+lever2.getName()+" being unpulled"); }
            locks(target,lever1);
            locks(target,lever2);
        };
        var releasedByEither = function(target, lever1, lever2) {
            if (!!!lever1.isPulled() && !!!lever2.isPulled()) { target.setLocked("Locked by both "+lever1.getName()+" and "+lever2.getName() +" being unpulled"); }
            locks(target,lever1);
            locks(target,lever2);
        };
        var locksN = function(lever, condition, target) {
            if (lever.isPulled() && !!!condition.isPulled()) { target.setLocked("Locked by "+lever.getName()+" being pulled and "+condition.getName()+" being unpulled"); }
        };
        var locksR = function(lever, condition, target) {
            if (lever.isPulled() && condition.isPulled()) { target.setLocked("Locked by both "+lever.getName()+" and "+condition.getName()+" being pulled"); }
        };
        var locksBW = locks;
        
        // Lever  1 Released by          Locks 11
        // Lever  2 Released by 7        Locks 15
        // Lever  3 Released by BLOCK    Locks 7(B/w), 10(B/W), 15, 19, 22
        // Lever  4 Released by 3, 5     Locks 
        // Lever  5 Released by          Locks 7, 10, 15
        // Lever  6 Released by          Locks (Lever 11 locks 6 electrically without a reverse lock applied (i.e. 6 does not lock 11)
        // Lever  7 Released by          Locks 5, 8, 10
        // Lever  8 Released by          Locks 7, 18
        // Lever  9 Released by          Locks 8(B/W)
        // Lever 10 Released by          Locks 5, 7, 12, 15, 16, 18
        // Lever 11 Released by          Locks 1, 6 (see lever 6)
        // Lever 12 Released by          Locks 10, 16, 18
        // Lever 13 Released by BLOCK    Locks 10(B/W), 12(B/W), (21 when 8R)
        // Lever 14 Released by          Locks 15, 16, 18, 20
        // Lever 15 Released by          Locks 2, 3, 5, 10, 14, 16, 18, 19, 20
        // Lever 16 Released by 8, 9     Locks 10, 12, 14, 15
        // Lever 17 Released by 11       Locks 10(B/W), 12(B/W) (21 when 8N) (also requires 17 free light lit)
        // Lever 18 Released by 9        Locks 8, 10, 12, 14, 15
        // Lever 19 Released by BLOCK    Locks 3, 7(B/W), 10(B/W), 15, 22
        // Lever 20 Released by 9        Locks 14, 15
        // Lever 21 Released by 10 or 12 Locks 8(B/W), (13 when 8R), (17 when 8N), 22, 23
        // Lever 22 Released by 9, 10    Locks 3, 19, 21
        // Lever 23 Released by 9, 12    Locks 21

        // Stage 0 assume unlocked
        this._lever.forEach(function(v,i,a) { a[i].unlock(); });
        
        // Stage 1 Apply released by rules
        releasedBySingle(this._lever[2],this._lever[7]);
        releasedByBoth(this._lever[4],this._lever[3],this._lever[5]);
        releasedByBoth(this._lever[16],this._lever[8],this._lever[9]);
        releasedBySingle(this._lever[17],this._lever[11]);
        releasedBySingle(this._lever[18],this._lever[9]);
        releasedBySingle(this._lever[20],this._lever[9]);
        releasedByEither(this._lever[21],this._lever[10],this._lever[12]);
        releasedByBoth(this._lever[22],this._lever[9],this._lever[10]);
        releasedByBoth(this._lever[23],this._lever[9],this._lever[12]);

        // Stage 2 standard locks - if the lever is pulled, the target must be held normal - note that the chart has the reciprocal locks
        // defined (so we don't need to check target is at normal - if it get pulled, lever must be held normal
        locks(this._lever[ 1],this._lever[11]);
        locks(this._lever[ 2],this._lever[15]);
        locks(this._lever[ 3],this._lever[15]);
        locks(this._lever[ 3],this._lever[19]);
        locks(this._lever[ 3],this._lever[22]);

        locks(this._lever[ 5],this._lever[ 7]);
        locks(this._lever[ 5],this._lever[10]);
        locks(this._lever[ 5],this._lever[15]);

        locks(this._lever[ 7],this._lever[ 5]);
        locks(this._lever[ 7],this._lever[ 8]);
        locks(this._lever[ 7],this._lever[10]);
        locks(this._lever[ 8],this._lever[ 7]);
        locks(this._lever[ 8],this._lever[18]);
        locks(this._lever[10],this._lever[ 5]);
        locks(this._lever[10],this._lever[ 7]);
        locks(this._lever[10],this._lever[12]);
        locks(this._lever[10],this._lever[15]);
        locks(this._lever[10],this._lever[16]);
        locks(this._lever[10],this._lever[18]);
        locks(this._lever[11],this._lever[ 1]);
        locks(this._lever[12],this._lever[10]);
        locks(this._lever[12],this._lever[16]);
        locks(this._lever[12],this._lever[18]);

        locks(this._lever[14],this._lever[15]);
        locks(this._lever[14],this._lever[16]);
        locks(this._lever[14],this._lever[16]);
        locks(this._lever[14],this._lever[18]);
        locks(this._lever[14],this._lever[20]);
        locks(this._lever[15],this._lever[ 2]);
        locks(this._lever[15],this._lever[ 3]);
        locks(this._lever[15],this._lever[ 5]);
        locks(this._lever[15],this._lever[10]);
        locks(this._lever[15],this._lever[14]);
        locks(this._lever[15],this._lever[16]);
        locks(this._lever[15],this._lever[18]);
        locks(this._lever[15],this._lever[19]);
        locks(this._lever[15],this._lever[20]);
        locks(this._lever[16],this._lever[10]);
        locks(this._lever[16],this._lever[12]);
        locks(this._lever[16],this._lever[14]);
        locks(this._lever[16],this._lever[15]);

        locks(this._lever[18],this._lever[ 8]);
        locks(this._lever[18],this._lever[10]);
        locks(this._lever[18],this._lever[12]);
        locks(this._lever[18],this._lever[14]);
        locks(this._lever[18],this._lever[15]);
        locks(this._lever[19],this._lever[ 3]);
        locks(this._lever[19],this._lever[15]);
        locks(this._lever[19],this._lever[22]);
        locks(this._lever[20],this._lever[14]);
        locks(this._lever[20],this._lever[15]);
        locks(this._lever[21],this._lever[22]);
        locks(this._lever[21],this._lever[23]);
        locks(this._lever[22],this._lever[ 3]);
        locks(this._lever[22],this._lever[19]);
        locks(this._lever[22],this._lever[21]);
        locks(this._lever[23],this._lever[21]);

        // Stage 3 N locks - if the lever if pulled and the conditional is not, the target must be held normal (again reciposity in the chart)
        locksN(this._lever[17],this._lever[8],this._lever[21]);
        locksN(this._lever[21],this._lever[8],this._lever[17]);

        // Stage 4 R locks - if the lever is pulled and the conditional is also pulled, the target must be held normal (reciprocal chart)
        locksR(this._lever[13],this._lever[8],this._lever[21]);
        locksR(this._lever[21],this._lever[8],this._lever[13]);

        // Stage 5 B/W locks - if the lever is pulled, the target is locked (implementation same as for standard locks - the lack of reciposity in
        // the chart is what gives the B/W bahaviour
        locksBW(this._lever[ 3],this._lever[ 7]);
        locksBW(this._lever[ 3],this._lever[10]);
        locksBW(this._lever[ 9],this._lever[ 8]);
        locksBW(this._lever[13],this._lever[10]);
        locksBW(this._lever[13],this._lever[12]);
        locksBW(this._lever[17],this._lever[10]);
        locksBW(this._lever[17],this._lever[12]);
        locksBW(this._lever[19],this._lever[ 7]);
        locksBW(this._lever[19],this._lever[10]);
        locksBW(this._lever[21],this._lever[ 8]);

        // Stage 6 Note 1 - Lever 11 locks 6 electrically without a reverse lock (i.e. 6 does not lock 11)
        locks(this._lever[11],this._lever[ 6]);

        // Stage 7 apply the absolute block instrument rules
        if (this._redbridgeUp.getState() !== CommuatorState.LINE_CLEAR && !!!this._lever[13].isPulled()) {
            this._lever[13].setLocked("Locked because Redbridge Up Line is not LINE CLEAR and this lever is not pulled");
        }
        if (this._kimbridgeDown.getState() !== CommuatorState.LINE_CLEAR && !!!this._lever[3].isPulled()) {
            this._lever[3].setLocked("Locked because Kimbridge Down Line is not LINE CLEAR and this lever is not pulled");
        }
        if (this._kimbridgeDown.getState() !== CommuatorState.TRAIN_ON_LINE && !!!this._lever[19].isPulled()) {
            this._lever[19].setLocked("Locked because Kimbridge Down Line is not TRAIN_ON_LINE and this lever is not pulled");
        }

        // Step 8 apply single line working rules
        if (!!!this._free17 && !!!this._lever[17].isPulled()) {
            this._lever[17].setLocked("Locked because the eastleigh box has not released the line and this lever is not pulled");
        }
    }
};