//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
/* globals TimeEvent:true, RealTimeEvent : true, GameTimeEvent:true, Capture, Subclass, */
/* jshint esnext: true, debug: true */
function TimeEvent(scheduler, what, who, ...why) {
    this._scheduler = scheduler;
    this._what = what;
    this._who = who;
    this._why = why;
    this._id = this._timeEventId++;
}
TimeEvent.prototype = {
    doEvent: function() {
            Capture.exception(this._scheduler._id,this._what,this._who, ...this._why);
    },
    isReady: function() {
        return (this.getRemainingMillis() < 0);
    },
    getRemainingMillis: function() {
        throw new Error("getRemainingTime not overriden");
    },
    cancel: function() {
        this._scheduler.cancelEvent(this);
    },
    isEarlierThan: function(other) {
        const ourRemainingTime = this.getRemainingMillis();
        const otherRemainingTime = other.getRemainingMillis();
        if (ourRemainingTime < otherRemainingTime) {
            return true;
        } else if (ourRemainingTime > otherRemainingTime) {
            return false;
        } else {
            // Same time, so the first registered wins
            return (this._id < other._id);
        }
    },
    _timeEventId: 1
};

function RealTimeEvent(scheduler, when, what, who, ...why) {
    TimeEvent.call(this,scheduler, what, who, ...why);
    this._when = when*1000 + Date.now();
}
Subclass.make(RealTimeEvent,TimeEvent);
RealTimeEvent.prototype.getRemainingMillis = function() {
    return (this._when - Date.now());
};
function GameTimeEvent(scheduler, when, what, who, ...why) {
    TimeEvent.call(this,scheduler, what, who, ...why);
    this._when = when + scheduler._clock.getTicks();
}
Subclass.make(GameTimeEvent,TimeEvent);
GameTimeEvent.prototype.getRemainingMillis = function() {
    return (this._when-this._scheduler._clock.getTicks())*1000/this._scheduler._warpFactor;
};

function Scheduler(id) {
    this._id = id;
    this._events = [];
    this._warpFactor = 1;
    this._clock = {getTicks: function() { return 0; }}; // Initialise to an 'null' clock
    this._eventTimer = null; // No timeout initially set
}

Scheduler.prototype = {
    setClock : function(clock) {
        this._clock = clock;
    },
    setWarpFactor: function(factor) {
        this._warpFactor = factor;
        this.scheduleNextEvent(); // This removes the currently scheduled next event and replaces it with the (now) next event
    },
    getWarpFactor: function() {
        return this._warpFactor;
    },
    isEmpty: function() {
        return this._events.length === 0;
    },
    executeEvents: function() {
        var doneAnEvent;
        do {
            doneAnEvent = this.executeFirstEvent();
        } while(doneAnEvent);
        this.scheduleNextEvent();
    },
    executeFirstEvent: function() {
        if (this.isEmpty()) {
            return false;
        }
        this._moveNextEventToFront();
        if (this._events[0].isReady()) {
            this._events.shift().doEvent();
            return true;
        } else {
            return false;
        }
    },
    _checkEventParams: function(when,what) {
        const isFunction = function(object) {
            return !!(object && object.constructor && object.call && object.apply);
        };
        const isNumber = function(n) {
            return !(isNaN(parseFloat(n)) && !isNaN(n - 0)); 
        };
        if (!!!isFunction(what)) {
            debugger;
            throw new Error("No function to call provided when adding an event");
        }
        if (!!!isNumber(when)) {
            debugger;
            throw new Error("No realOffset provided when adding an event");
        }
    },
    addRealTimeEvent: function(when,what,who,...why) {
        return this._addEvent(function(scheduler,when,what,who,...why) { return new RealTimeEvent(scheduler,when,what,who,...why); },
                when,what,who,...why);
    },
    addGameTimeEvent: function(when,what,who,...why) {
        return this._addEvent(function(scheduler,when,what,who,...why) { return new GameTimeEvent(scheduler,when,what,who,...why); },
                when,what,who,...why);
    },
    _addEvent: function(factory,when,what,who,...why){
        this._checkEventParams(when,what);
        const event = factory(this,when,what,who,...why);
        this._events.push(event);
        this.scheduleNextEvent();
        return event;
    },
    cancelEvent: function(event) {
        for(var i=0; i<this._events.length; i++) {
            if (this._events[i] === event) {
                this._events.splice(i,1);
                break;
            }
        }
        this.scheduleNextEvent();
    },
    _moveNextEventToFront: function() {
        var earliestEventIndex = 0; 
        for (var i=1; i<this._events.length; i++) {
            if (this._events[i].isEarlierThan(this._events[earliestEventIndex])) {
                earliestEventIndex = i;
            }
        }
        if (earliestEventIndex !== 0) {
            const temp = this._events[0];
            this._events[0] = this._events[earliestEventIndex];
            this._events[earliestEventIndex] = temp;
        }
    },
    scheduleNextEvent: function() {
        if (this._eventTimer) {
            window.clearTimeout(this._eventTimer);
            this._eventTimer = null;
        }
        if (!!!this.isEmpty()) {
            this._moveNextEventToFront();
            const when = (this._events[0].getRemainingMillis()<0 ? 0 : this._events[0].getRemainingMillis());
            const us = this; // Capture this
            this._eventTimer = window.setTimeout(function() { us.executeEvents(); },when);
        }
    }
};