//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
/* globals BoxState : true, BellCode, CommuatorState, Subclass */
/* jshint esnext: true */
function BoxState(description) {
    this._description = description;
}
BoxState.prototype = {
    getTooltip: function() { return this._description; }
};

BoxState.NONE = new BoxState("Idle");
BoxState.RECEIVING = new BoxState("Expecting a bell code");
BoxState.AWAITING_TRAIN_ENTERING = new BoxState("Waiting for a 'Train Entering Section' bell code");
BoxState.AWAITING_ATTN_ACK = new BoxState("Waiting for the acknowledgement of an 'Attention' bell code (so they can offer a train)");
BoxState.AWAITING_OFFER_ACK = new BoxState("Waiting for an acknowledgment of their offer of a train");
BoxState.AWAITING_THEIR_LINE_CLEAR = new BoxState("Waiting for the line to Romsey to be LINE CLEAR");
BoxState.PAUSE_TRAIN_START = new BoxState("Arranging for a train to signalled through to Romsey");
BoxState.AWAITING_ENTERING_ACK = new BoxState("Waiting for an acknowledgement of 'Train Entering Section' bell code");
BoxState.AWAITING_THEIR_TRAIN_ON_LINE = new BoxState("Waiting for the line to Romsey to be TRAIN_ON_LINE");
BoxState.AWAITING_EXIT_ATTN_ACK = new BoxState("Waiting for an acknowledgement of an ' Attention' bell code (so they can send 'Train Out Of Section')");
BoxState.AWAITING_OUT_OF_SECTION_ACK = new BoxState("Waiting for an acknowledgement of their 'Train Out Of Section' bell code");
BoxState.AWAITING_SHUNT_ATTN = new BoxState("Waiting for an attn that will signal end of 'Shunt' bell code");
BoxState.AWAITING_END_SHUNT = new BoxState("Waiting for an 'End Shunt' bell code");

function SignalBox(controller, bell) {
    this._controller = controller;
    this._scheduler = controller.getScheduler();
    this._name = bell.getName();
    this._bell = bell;
    this._state = BoxState.NONE;
    this._preOfferState = BoxState.NONE;
    this._bell.setTooltip(this._state.getTooltip());
}

SignalBox.prototype = {
    changeState: function(state) {
        if (state !== this._state) {
            this._state = state;
            this._bell.setTooltip(state.getTooltip());
        }
    },
    changePrevState: function(state) {
        this._prevState = state;
    },
    _scheduleTimeoutAck: function() {
        this._scheduler.addGameTimeEvent(60,this.timeoutAck,this,this._state);
    },
    timeoutAck: function(oldState) {
        if (oldState === this._state) {
            if (this._bell.isInBellCode()) {
                this._scheduler.addRealTimeEvent(5,this.timeoutAck,this,this._state); // Give them some (real) time to complete their tapping
            } else {
                var reschedule = false;
                var bellcode = null;
                // The state hasn't changed, so we haven't had the expected ack - we need to repeat our bell code
                switch(oldState) {
                    case BoxState.AWAITING_OFFER_ACK:
                    case BoxState.AWAITING_THEIR_LINE_CLEAR:
                        this.changeState(BoxState.AWAITING_ATTN_ACK);
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_ATTN_ACK:
                    case BoxState.AWAITING_EXIT_ATTN_ACK:
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_ENTERING_ACK:
                        bellcode = BellCode.EnteringSection;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_THEIR_TRAIN_ON_LINE:
                        this.changeState(BoxState.AWAITING_ENTERING_ACK);
                        bellcode = BellCode.EnteringSection;
                        reschedule = true;
                        break;
                    case BoxState.AWAITING_OUT_OF_SECTION_ACK:
                        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
                        bellcode = BellCode.Attn;
                        reschedule = true;
                        break;
                    default:
                        // Unexpected, but we'll try to recover to by just ignoring this timeout event
                        break;
                }
                if (bellcode) {
                    this._bell.scheduleBellCode(bellcode,0);
                }
                if (reschedule) {
                    this._scheduleTimeoutAck();
                }
            }
        }
    },
    sendOutOfOurSection: function() {
        if (this._state === BoxState.PAUSE_TRAIN_START) {
            this._bell.scheduleBellCode(BellCode.EnteringSection,5);
            this.changeState(BoxState.AWAITING_ENTERING_ACK);
            this._scheduleTimeoutAck();
        }
    },
    getName: function() {
        return this._name;
    },
    getBell: function() {
        return this._bell;
    },
    getState: function() {
        return this._state;
    },
    scheduleBellCode: function(code,offset) {
        this._bell.scheduleBellCode(code,offset);
    },
    offerTrain: function(code,whatToCallOnceTrainOnLine,whoToCallOnceTrainOnLine,whyToProvideOnceTrainOnLine) {
        if (this.canTrainBeOffered()) { // Note: this method must be provided by the actual signal box objects once their construction is complete
            this._preOfferState = this._state;
            this._state = BoxState.AWAITING_ATTN_ACK;
            this._trainToOffer = code;
            this._whatToCallOnceTrainOnLine = whatToCallOnceTrainOnLine;
            this._whoToCallOnceTrainOnLine = whoToCallOnceTrainOnLine;
            this._whyToProvideOnceTrainOnLine = whyToProvideOnceTrainOnLine;
            this._bell.scheduleBellCode(BellCode.Attn,1);
            this._scheduleTimeoutAck();
            return true;
        } else {
            return false;
        }
    },
    makeTrainOnLineCall: function() {
        this._scheduler.addRealTimeEvent(0.1,this._whatToCallOnceTrainOnLine,this._whoToCallOnceTrainOnLine,this.whyToProvideOnceTrainOnLine);
    },
    actOnBellPush: function(pushedBellCode) {
        throw new TypeError("actOnBellCode not overridden");
    },
    exitTrain: function() {
        throw new TypeError("exitTrain not overridden");
    },
    canTrainBeOffered: function() {
        throw new TypeError("canTrainBeOffered not overridden");
    }
};

function SingleLineBox(controller, bell, model) {
    SignalBox.call(this,controller, bell);
    this._model = model;
}
Subclass.make(SingleLineBox,SignalBox);

SingleLineBox.prototype.checkDirectionLever = function(need) {
    if (need && this._state === BoxState.AWAITING_THEIR_LINE_CLEAR && this._model.getLever(6).isPulled()) {
        this.changeState(BoxState.PAUSE_TRAIN_START);
        this._scheduler.addGameTimeEvent(60,this.sendOutOfOurSection,this);
    } else if (!!!need && this._state === BoxState.AWAITING_THEIR_TRAIN_ON_LINE && !!!this._model.getLever(6).isPulled()) {
        this.makeTrainOnLineCall();
        this._scheduler.addRealTimeEvent(1,this.changeState,this,this._preOfferState);
    } else {
        this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,need);
    }
};
SingleLineBox.prototype.exitTrain = function() {
    if (this._state !== BoxState.NONE) {
        return false;
    }
    if (!!!this._model.is17Free()) {
        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
        const bell_length = this._bell.scheduleBellCode(BellCode.Attn,1);
        this._scheduleTimeoutAck();
        return true;
    } else {
        return false;
    }
};
SingleLineBox.prototype.canTrainBeOffered = function() {
    return this._state === BoxState.NONE;
};
SingleLineBox.prototype.actOnBellPush = function() {
    switch(this._state) {
        case BoxState.NONE:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.RECEIVING);
            }
            break;
        case BoxState.RECEIVING:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode()) || BellCode.Cancel.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.NONE);
/* jshint ignore:start */
            } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                // Ignore Bell Code
            } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.release17,this._model);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_SHUNT_ATTN);
            } else {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.release17,this._model);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_TRAIN_ENTERING);
            }
            break;
        case BoxState.AWAITING_TRAIN_ENTERING:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode()) || 
                BellCode.WronglyDescribed.equals(this._bellCode.getBellCode()) ||
                BellCode.Cancel.equals(this._bellCode.getBellCode())
               ) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.lock17,this._model);
                var newState = BoxState.NONE;
                if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) { newState = BoxState.RECEIVING; }
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,newState);
            }
            break;
        case BoxState.AWAITING_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OFFER_ACK);
                const bellLength = this._bell.scheduleBellCode(this._trainToOffer,1);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OFFER_ACK:
            if (this._trainToOffer.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_LINE_CLEAR);
                this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,true);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_LINE_CLEAR:
            break;
        case BoxState.PAUSE_TRAIN_START:
            break;
        case BoxState.AWAITING_ENTERING_ACK:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(0.1,this.checkDirectionLever,this,false);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_TRAIN_ON_LINE:
            break;
        case BoxState.AWAITING_EXIT_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OUT_OF_SECTION_ACK);
                const bellLength = this._bell.scheduleBellCode(BellCode.OutOfSection,1);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OUT_OF_SECTION_ACK:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                this._scheduler.addRealTimeEvent(1.5,this.changeState,this,BoxState.NONE);
            }
            break;
        case BoxState.AWAITING_SHUNT_ATTN:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.AWAITING_END_SHUNT);
            }
            break;
        case BoxState.AWAITING_END_SHUNT:
            if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._model.lock17,this._model);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
            }
            break;
        default:
            break;
    }
};

function DoubleLineBox(controller, bell, ourLine, theirLine) {
    SignalBox.call(this,controller, bell);
    this._ourLine = ourLine;
    this._theirLine = theirLine;
}
Subclass.make(DoubleLineBox,SignalBox);

DoubleLineBox.prototype.checkTheirCommuator = function(need) {
    if (need === CommuatorState.LINE_CLEAR && this._state === BoxState.AWAITING_THEIR_LINE_CLEAR && this._theirLine.getState() === need) {
        this.changeState(BoxState.PAUSE_TRAIN_START);
        this._scheduler.addGameTimeEvent(60,this.sendOutOfOurSection,this);
    } else if (need === CommuatorState.TRAIN_ON_LINE && this._state === BoxState.AWAITING_THEIR_TRAIN_ON_LINE && this._theirLine.getState() === need) {
        this.makeTrainOnLineCall();
        this._scheduler.addRealTimeEvent(1,this.changeState,this,this._preOfferState);
    } else {
        this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,need);
    }
};
DoubleLineBox.prototype.exitTrain = function() {
    if (this._state !== BoxState.NONE) {
        return false;
    }
    if (this._ourLine.getState() === CommuatorState.TRAIN_ON_LINE) {
        this.changeState(BoxState.AWAITING_EXIT_ATTN_ACK);
        const bell_length = this._bell.scheduleBellCode(BellCode.Attn,1);
        this._scheduleTimeoutAck();
        return true;
    } else {
        return false;
    }
    
};
DoubleLineBox.prototype.canTrainBeOffered = function() {
    return this._state === BoxState.NONE || this._state === BoxState.AWAITING_TRAIN_ENTERING || this._state === BoxState.AWAITING_SHUNT_ATTN;
};
DoubleLineBox.prototype.actOnBellPush = function() {
    switch(this._state) {
        case BoxState.NONE:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.RECEIVING);
            }
            break;
        case BoxState.RECEIVING:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode()) || BellCode.Cancel.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.NONE);
/* jshint ignore:start */
            } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                // Ignore Bell Code
            } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_SHUNT_ATTN);
            } else {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.LINE_CLEAR);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.AWAITING_TRAIN_ENTERING);
            }
            break;
        case BoxState.AWAITING_TRAIN_ENTERING:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode()) || 
                    BellCode.WronglyDescribed.equals(this._bellCode.getBellCode()) ||
                    BellCode.Cancel.equals(this._bellCode.getBellCode())
                   ) {
                const bellLength = this._bell.echoBellCode();
                if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                    this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                } else {
                    this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                }
                if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) {
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.RECEIVING);
                } else {
                    this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
                }
            }
            break;
        case BoxState.AWAITING_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OFFER_ACK);
                this._bell.scheduleBellCode(this._trainToOffer,0);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OFFER_ACK:
            if (this._trainToOffer.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_LINE_CLEAR);
                this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,CommuatorState.LINE_CLEAR);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_THEIR_LINE_CLEAR:
            break;
        case BoxState.PAUSE_TRAIN_START:
            // We're due to send them an "Entering Section" bell, but they want to something in the meantime
            switch(this._preOfferState) {
                case BoxState.NONE:
                    if (BellCode.Attn.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this._preOfferState = BoxState.RECEIVING;
                    }
                    break;
                case BoxState.RECEIVING:
                    if (BellCode.OutOfSection.equals(this._bell.getBellCode()) || BellCode.Cancel.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this._preOfferState = BoxState.NONE;
/* jshint ignore:start */
                    } else if (BellCode.Attn.equals(this._bell.getBellCode()) || BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                        // Ignore Bell Code
                    } else if (BellCode.Shunt.equals(this._bell.getBellCode())) {
/* jshint ignore:end */
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.AWAITING_SHUNT_ATTN);
                    } else {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.LINE_CLEAR);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.AWAITING_TRAIN_ENTERING);
                    }
                    break;
                case BoxState.AWAITING_TRAIN_ENTERING:
                    if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.TRAIN_ON_LINE);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.NONE);
                    } else if (BellCode.WronglyDescribed.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.RECEIVING);
                    }
                    break;
                case BoxState.AWAITING_SHUNT_ATTN:
                    if (BellCode.Attn.equals(this._bell.getBellCode())) {
                        this._bell.echoBellCode();
                        this._preOfferState = BoxState.AWAITING_END_SHUNT;
                    }
                    break;
                case BoxState.AWAITING_END_SHUNT:
                    if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                        const bellLength = this._bell.echoBellCode();
                        this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                        this._scheduler.addRealTimeEvent(bellLength+1.01,this.changePrevState,this,BoxState.NONE);
                    }
                    break;
                default:
                    break;
            }
            break;
        case BoxState.AWAITING_ENTERING_ACK:
            if (BellCode.EnteringSection.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_THEIR_TRAIN_ON_LINE);
                this._scheduler.addRealTimeEvent(0.1,this.checkTheirCommuator,this,CommuatorState.TRAIN_ON_LINE);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_EXIT_ATTN_ACK:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this.changeState(BoxState.AWAITING_OUT_OF_SECTION_ACK);
                const bellLength = this._bell.scheduleBellCode(BellCode.OutOfSection,0);
                this._scheduleTimeoutAck();
            }
            break;
        case BoxState.AWAITING_OUT_OF_SECTION_ACK:
            if (BellCode.OutOfSection.equals(this._bell.getBellCode())) {
                this._scheduler.addRealTimeEvent(1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                this._scheduler.addRealTimeEvent(1.01,this.changeState,this,BoxState.NONE);
            }
            break;
        case BoxState.AWAITING_SHUNT_ATTN:
            if (BellCode.Attn.equals(this._bell.getBellCode())) {
                this._bell.echoBellCode();
                this.changeState(BoxState.AWAITING_END_SHUNT);
            }
            break;
        case BoxState.AWAITING_END_SHUNT:
            if (BellCode.EndShunt.equals(this._bell.getBellCode())) {
                const bellLength = this._bell.echoBellCode();
                this._scheduler.addRealTimeEvent(bellLength+1,this._ourLine.setState,this._ourLine,CommuatorState.NORMAL);
                this._scheduler.addRealTimeEvent(bellLength+1.01,this.changeState,this,BoxState.NONE);
            }
            break;
        default:
            break;
    }
};