//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

/* globals StartTrainAction, PauseTrainAction, UpdateLocationTrainAction, WaitAtSignalTrainAction, StopTrainAction, JustDoneTrainAction */
/* globals SetEastleighSignalsTrainAction, SetEastleighPointTrainAction */
/* jshint esnext:true */
function Train(scheduler,name,actions) {
    this._scheduler = scheduler;
    this._name = name;
    this._actions = actions;
    this._location = null;
    this._nextStep = 0;
    this._listeners = [];
}
Train.prototype = {
    addListener: function(listener) {
        this._listeners.push(listener);
    },
    doneArrival: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i].arrival();
        }
        this.doneStep();
    },
    doneDeparture: function() {
        for(var i = 0; i<this._listeners.length; i++) {
            this._listeners[i].departure();
        }
        this.doneStep();
    },
    getName: function() { return this._name; },
    setLocation: function(location) { this._location = location; },
    getLocation: function() { return this._location; },
    isActive: function() { return (this._nextStep < this._actions.length); },
    doNextStep: function() {
        if (this.isActive()) {
            const action = this._actions[this._nextStep];
            var done = this.doneStep;
            if (action instanceof JustDoneTrainAction) {
                if (action.getKind() === "arrival") {
                    done = this.doneArrival;
                } else if (action.getKind() === "depart") {
                    done = this.doneDeparture;
                }
            }
            this._actions[this._nextStep].doAction(this._scheduler,done,this,null);
        }
    },
    doneStep: function() {
        this._nextStep++;
        this.doNextStep();
    }
};

function TrainFactory(scheduler,model,name) {
    this._scheduler = scheduler;
    this._model = model;
    this._name = name;
    this._actions = [];
}
TrainFactory.prototype = {
    getName: function() { return this._name; },
    make: function() {
        return new Train(this._scheduler,this._name,this._actions);
    },
    setActions: function(actions) { this._actions = actions; },
    startTrain: function(box,bellcode) {
        return new StartTrainAction(box,bellcode);
    },
    pause: function(duration) {
        return new PauseTrainAction(duration);
    },
    updateLocation: function(pos) {
        const newPosition = (pos ? this._model.getPosition(pos) : pos);
        const result = new UpdateLocationTrainAction(this._location,newPosition);
        this._location = newPosition;
        return result;
    },
    waitAtSignal: function(lever) {
        const signalLever = this._model.getLever(lever);
        return new WaitAtSignalTrainAction(signalLever);
    },
    setEastleighSignals: function(first,next) {
        return new SetEastleighSignalsTrainAction(this._model,first,next);
    },
    setEastleighPoint: function(state) {
        return new SetEastleighPointTrainAction(this._model,state);
    },
    stopTrain: function(box) {
        return new StopTrainAction(box);
    },
    arrived: function() {
        return new JustDoneTrainAction("arrival");
    },
    depart: function() {
        return new JustDoneTrainAction("depart");
    }
};