//
// © Copyright 2020 David Vines
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived 
//    from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
/* globals SignalBox, Capture, Subclass */
/* jshint esnext: true */

// The basic TrainAction
function TrainAction(){}
TrainAction.prototype = {
    doAction: function(scheduler,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
        const done = this.tryAction(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
        
        if (done.retry) {
            if (done.realTime !== undefined) {
                scheduler.addRealTimeEvent(done.realTime,this.doAction,this,scheduler,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
            } else {
                scheduler.addGameTimeEvent(done.gameTime,this.doAction,this,scheduler,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
            }
        } else {
            if (done.scheduleDone) {
                if (done.realTime === undefined && done.gameTime === undefined) {
                    Capture.exception(scheduler,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
                } else {
                    if (done.gameTime === undefined) {
                        scheduler.addRealTimeEvent(done.realTime,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
                    } else {
                        scheduler.addGameTimeEvent(done.gameTime,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
                    }
                }
            }
        }
    },
    tryAction: function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
        throw new TypeError("TrainAction.tryAction not overriden");
    }
};

// Start a new train
function StartTrainAction(startLocation,bellcode) {
    TrainAction.call(this);
    this._startLocation = startLocation;
    this._bellcode = bellcode;
}
Subclass.make(StartTrainAction,TrainAction);

StartTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    if (this._startLocation === "Romsey") {
        // Easy, it's here!
        return { retry: false, scheduleDone: true };
    }
    if (this._startLocation instanceof SignalBox) {
        var accepted = this._startLocation.offerTrain(this._bellcode,whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone);
        if (accepted) {
            return { retry: false, scheduleDone: false };
        } else {
            return { retry: true, realTime: 1 };
        }
    }
    throw new Error("_startLocation of "+this._startLocation+" not recognised when attempting to start a train");
};

// Pause for a fixed period of time
function PauseTrainAction(duration) {
    TrainAction.call(this);
    this._duration = duration;
}
Subclass.make(PauseTrainAction,TrainAction);

PauseTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    return { retry:false, scheduleDone: true, gameTime: this._duration };
};

// Update the location of a train
function UpdateLocationTrainAction(oldPosition,position) {
    TrainAction.call(this);
    this._oldPosition = oldPosition;
    this._position = position;
}
Subclass.make(UpdateLocationTrainAction,TrainAction);

UpdateLocationTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    if (this._oldPosition) {
        this._oldPosition.setState(false);
    }
    if (this._position) {
        this._position.setState(true);
    }
    return { retry: false, scheduleDone: true, realTime: 0 };
};

// Wait at signal
function WaitAtSignalTrainAction(signal) {
    TrainAction.call(this);
    this._signal = signal;
}
Subclass.make(WaitAtSignalTrainAction,TrainAction);

WaitAtSignalTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    if (this._signal.isPulled()) {
        return { retry: false, scheduleDone: true, realTime: 0};
    } else {
        return { retry: true, scheduleDone: false, realTime: 0.1 };
    }
};

// Set Eastleigh signals
function SetEastleighSignalsTrainAction(model,zw30,next) {
    TrainAction.call(this);
    this._model = model;
    this._zw30 = zw30;
    this._next = next;
}
Subclass.make(SetEastleighSignalsTrainAction,TrainAction);

SetEastleighSignalsTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    this._model.getEastleigh().setSignals(this._zw30,this._next);
    return { retry: false, scheduleDone: true, realTime: 0 };
};

//Set Eastleigh point
function SetEastleighPointTrainAction(model,state) {
    TrainAction.call(this);
    this._model = model;
    this._state = state;
}
Subclass.make(SetEastleighPointTrainAction,TrainAction);

SetEastleighPointTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    this._model.getEastleigh().setPoint(this._state);
    return { retry: false, scheduleDone: true, realTime: 0 };
};

// Stop Train
function StopTrainAction(location) {
    TrainAction.call(this);
    this._location = location;
}
Subclass.make(StopTrainAction,TrainAction);

StopTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    if (this._location === "Romsey") {
        // Easy, it's here!
        return { retry: false, scheduleDone: false };
    }
    if (this._location instanceof SignalBox) {
        var accepted = this._location.exitTrain();
        if (accepted) {
            return { retry: false, scheduleDone: false };
        } else {
            return { retry: true, realTime: 1 };
        }
    }
    throw new Error("_location of "+this._location+" not recognised when attempting to stop a train");

};

function JustDoneTrainAction(kind) {
    TrainAction.call(this);
    this._kind = kind;
}
Subclass.make(JustDoneTrainAction,TrainAction);
JustDoneTrainAction.prototype.tryAction = function(whatToCallOnceDone,whoToCallOnceDone,whyToProvideOnceDone) {
    return { retry: false, scheduleDone: true, realTime: 0, gameTime: 0, };
};
JustDoneTrainAction.prototype.getKind = function() { return this._kind; };